//
// Created by robin on 10/8/17.
//

#include <libnfc_cpptools/StringUtils.h>
#include "libnfc_cpptools/FreeFareTagClassic.h"
#include "FreeFareTagClassicInternal.h"

namespace LibNfc
{
namespace FreeFare
{

FreeFareTagClassic::FreeFareTagClassic(std::shared_ptr<FreeFareTagClassicInternal> tag)
        : FreeFareTag(tag)
{
}

LibNfc::Utils::ResultBool FreeFareTagClassic::authenticate(int sector, const std::string& key, int keyType)
{
    return std::static_pointer_cast<FreeFareTagClassicInternal>(_tag)->authenticate(sector, key, keyType);
}

LibNfc::Utils::Result<MappedKeys> FreeFareTagClassic::mapKeys(const std::vector<std::string>& keys, std::function<void(int, int)> cb)
{
    MappedKeys mappedKeys;
    unsigned long done = 0;
    unsigned long total = 16 * keys.size();

    for (int i = 0; i < 16; ++i) {
        std::pair<std::string, std::string> blockKeys;
        for (const auto& key : keys)
        {
            if (authenticate(i, key, MFC_KEY_A)) {
                blockKeys.first = key;
            }
            if (authenticate(i, key, MFC_KEY_B)) {
                blockKeys.second = key;
            }
            if (cb != nullptr) {
                cb(++done, total);
            }
            if (!blockKeys.first.empty() && !blockKeys.second.empty()) {
                break;
            }
        }
        mappedKeys.push_back(blockKeys);
    }
    if (cb != nullptr && done < total) {
        cb(total, total);
    }

    return LibNfc::Utils::Result<MappedKeys>::ok(mappedKeys);
}

LibNfc::Utils::ResultString FreeFareTagClassic::readBlock(int sector, int block, const std::string& key, int keyType)
{
    return std::static_pointer_cast<FreeFareTagClassicInternal>(_tag)->readBlock(sector, block, key, keyType);
}

LibNfc::Utils::Result<FreeFareClassicSector> FreeFareTagClassic::readSector(int sector, const std::string& key, int keyType)
{
    std::string res;
    int lastBlock = std::static_pointer_cast<FreeFareTagClassicInternal>(_tag)->getSectorBlockCount(sector);
    for (int i = 0; i < lastBlock; ++i) {
        auto data = readBlock(sector, i, key, keyType);
        if (data) {
            res += data.getData();
        }
        else {
            return LibNfc::Utils::Result<FreeFareClassicSector>::error(data);
        }
    }
    return LibNfc::Utils::Result<FreeFareClassicSector>::ok(FreeFareClassicSector(res));
}

LibNfc::Utils::Result<std::vector<FreeFareClassicSector>> FreeFareTagClassic::read(const MappedKeys& keys, std::function<void(int, int)> cb)
{
    if (keys.size() != 16) {
        return LibNfc::Utils::Result<std::vector<FreeFareClassicSector>>::error("Must have 16 sectors keys");
    }
    int done = 0;
    int total = 4 * keys.size();
    std::vector<FreeFareClassicSector> sectors;
    for (int s = 0; s < keys.size(); ++s) {
        auto sectorKey = keys[s];
        FreeFareClassicSector sector;
        bool keyA = false;
        bool keyB = false;
        for (int b = 0; b < 3; ++b) {
            std::string data = "";
            if (!sectorKey.first.empty()) {
                auto blockResult = readBlock(s, b, sectorKey.first, MFC_KEY_A);
                if (blockResult) {
                    data = blockResult.getData();
                    keyA = true;
                }
            }
            if (!sectorKey.second.empty()) {
                auto blockResult = readBlock(s, b, sectorKey.second, MFC_KEY_B);
                if (blockResult) {
                    if (data.empty()) {
                        data = blockResult.getData();
                    }
                    keyB = true;
                }
            }
            sector.setBlock(b, data);
            if (cb != nullptr) {
                cb(++done, total);
            }
        }
        int b = 3;
        std::string dataA;
        std::string dataB;
        if (!sectorKey.first.empty()) {
            auto blockResult = readBlock(s, b, sectorKey.first, MFC_KEY_A);
            if (blockResult) {
                dataA = blockResult.getData();
                keyA = true;
            }
        }
        if (!sectorKey.second.empty()) {
            auto blockResult = readBlock(s, b, sectorKey.second, MFC_KEY_B);
            if (blockResult) {
                dataB = blockResult.getData();
                keyB = true;
            }
        }
        if (cb != nullptr) {
            cb(++done, total);
        }

        sector.setBlock(b, dataA);
        FreeFareClassicAccessBits accessBitsDboA = sector.getAccessBitsDbo();
        sector.setBlock(b, dataB);
        FreeFareClassicAccessBits accessBitsDboB = sector.getAccessBitsDbo();
        sector.setKeyA(keyA ? sectorKey.first : "");
        sector.setKeyB(keyB ? sectorKey.second : "");

        std::string accessBits;
        if (keyA && accessBitsDboA.canKeyAReadAccessBitsTrailer()) {
            accessBits = accessBitsDboA.getBits();
        }
        else if (keyB && accessBitsDboB.canKeyBReadAccessBitsTrailer()) {
            accessBits = accessBitsDboB.getBits();
        }
        sector.setAccessBits(accessBits);

        sectors.push_back(sector);
    }
    if (cb != nullptr && done < total) {
        cb(total, total);
    }

    return LibNfc::Utils::Result<std::vector<FreeFareClassicSector>>::ok(sectors);
}

LibNfc::Utils::Result<std::vector<FreeFareClassicSector>> FreeFareTagClassic::read(const std::vector<std::string>& keys,
                                                                     std::function<void(int, int)> mapCb,
                                                                     std::function<void(int, int)> readCb)
{
    auto mappedKeysResult = mapKeys(keys, mapCb);
    if (!mappedKeysResult) {
        return LibNfc::Utils::Result<std::vector<FreeFareClassicSector>>::error(mappedKeysResult);
    }
    return read(mappedKeysResult.getData(), readCb);
}

LibNfc::Utils::ResultBool FreeFareTagClassic::writeBlock(int sector, int block, const std::string& key, int keyType, const std::string& data)
{
    return std::static_pointer_cast<FreeFareTagClassicInternal>(_tag)->writeBlock(sector, block, key, keyType, LibNfc::Utils::StringUtils::ensureSize(data, 16));
}

LibNfc::Utils::ResultBool FreeFareTagClassic::writeSector(int sector, const std::string& key, int keyType, const std::string& data)
{
    std::string d = LibNfc::Utils::StringUtils::ensureSize(data, 64);
    std::string errors;
    for (int b = 0; b < 4; ++b) {
        auto result = writeBlock(sector, b, key, keyType, d.substr(b * 16, 16));
        if (!result) {
            errors += std::string(errors.empty() ? "" : "\n") + "Failed to write sector " + std::to_string(sector) +
                      " block " + std::to_string(b) + " with key " + (keyType == MFC_KEY_A ? "A" : "B") + ": " + result.getError();
        }
    }
    if (errors.empty()) {
        return LibNfc::Utils::ResultBool::ok(true);
    }
    return LibNfc::Utils::ResultBool::error(errors);
}

LibNfc::Utils::ResultBool FreeFareTagClassic::write(const MappedKeys& keys,
                                             const std::string& data,
                                             bool writeSector0,
                                             std::function<void(int, int)> cb)
{
    if (keys.size() != 16) {
        return LibNfc::Utils::ResultBool::error("Must have 16 sectors keys");
    }
    std::string d = LibNfc::Utils::StringUtils::ensureSize(data, 1024);
    std::string errors;
    unsigned long done = 0;
    unsigned long total = 4 * keys.size();
    for (unsigned long s = 0; s < keys.size(); ++s) {
        auto sectorKey = keys[s];
        for (int b = 0; b < 4; ++b) {
            if (s == 0 && b == 0 && !writeSector0) {
                continue;
            }
            std::string blockData = d.substr((s * 64) + (b * 16), 16);
            if (cb != nullptr && done < total) {
                bool keyA = false;
                bool keyB = false;
                std::string sectorErrors;
                if (!sectorKey.first.empty()) {
                    auto resultA = writeBlock(s, b, sectorKey.first, MFC_KEY_A, blockData);
                    if (resultA) {
                        keyA = true;
                    }
                    else {
                        sectorErrors = "Failed to write sector " + std::to_string(s) +
                                       " block " + std::to_string(b) + " with key A: " + resultA.getError();
                    }
                }
                if (!keyA && !sectorKey.second.empty()) {
                    auto resultB = writeBlock(s, b, sectorKey.first, MFC_KEY_A, blockData);
                    if (resultB) {
                        keyB = true;
                    }
                    else {
                        sectorErrors = std::string(sectorErrors.empty() ? "" : "\n") + "Failed to write sector " + std::to_string(s) +
                                       " block " + std::to_string(b) + " with key B: " + resultB.getError();
                    }
                }
                if (!keyA && !keyB) {
                    errors += std::string(errors.empty() ? "" : "\n") + sectorErrors;
                }
                cb(++done, total);
            }
        }
    }
    if (cb != nullptr && done < total) {
        cb(total, total);
    }
    if (errors.empty()) {
        return LibNfc::Utils::ResultBool::ok(true);
    }
    return LibNfc::Utils::ResultBool::error(errors);
}

LibNfc::Utils::ResultBool FreeFareTagClassic::write(const std::vector<std::string>& keys, const std::string& data, bool writeSector0,
                                             std::function<void(int, int)> mapCb, std::function<void(int, int)> writeCb)
{
    auto mappedKeysResult = mapKeys(keys, mapCb);
    if (!mappedKeysResult) {
        return LibNfc::Utils::ResultBool::error(mappedKeysResult);
    }
    return write(mappedKeysResult.getData(), data, writeSector0, writeCb);
}

}; // FreeFare
}; // LibNfc